{ Main view, where most of the application logic takes place.

  Feel free to use this code as a starting point for your own projects.
  This template code is in public domain, unlike most other CGE code which
  is covered by BSD or LGPL (see https://castle-engine.io/license). }
unit GameViewMain;

{$MODE DELPHI}

interface

uses Classes,
  CastleVectors, CastleComponentSerialize,
  CastleUIControls, CastleControls, CastleKeysMouse, CastleScene, castletransform, castleviewport, castlecolors,
  x3dnodes, castleimages, CastleShapes, castlecameras;

type
  { Main view, where most of the application logic takes place. }
  TViewMain = class(TCastleView)
  published
    { Components designed using CGE editor.
      These fields will be automatically initialized at Start. }
    LabelFps: TCastleLabel;
    Viewport1: TCastleViewport;
    ExamineNavigation1: TCastleExamineNavigation;    
  public
    constructor Create(AOwner: TComponent); override;
    procedure Start; override;
    procedure Update(const SecondsPassed: Single; var HandleInput: Boolean); override;
    function Press(const Event: TInputPressRelease): Boolean; override;
    function Motion(const Event: TInputMotion): Boolean; override;
  private
    FScene: TCastleScene;
    FMousePos: TVector2;
    function GetSceneUnderMouse(): TCastleScene;
    function GetSceneUnderMouseWithPixelTest(): TCastleScene;
    function GetSceneUnderMouseWithPixelTestByBugan(): TCastleScene;
  end;

var
  ViewMain: TViewMain;

implementation

uses SysUtils, generics.collections;

{ TViewMain ----------------------------------------------------------------- }

function IIF(ACond: Boolean; IfTrue, IfFalse: string): string;
begin
  if ACond then
    Result := IfTrue  
  else
    Result := IfFalse;
end;

constructor TViewMain.Create(AOwner: TComponent);
begin
  inherited;
  DesignUrl := 'castle-data:/gameviewmain.castle-user-interface';
end;

procedure TViewMain.Start;
begin
  inherited;
  ExamineNavigation1.TurnTable := True;
end;

function TViewMain.Motion(const Event: TInputMotion): Boolean; 
begin
  FMousePos := Event.Position;
  Result := inherited;
end;

procedure TViewMain.Update(const SecondsPassed: Single; var HandleInput: Boolean);
var LScene: TCastleScene;
begin
  inherited;
  { This virtual method is executed every frame (many times per second). }
  Assert(LabelFps <> nil, 'If you remove LabelFps from the design, remember to remove also the assignment "LabelFps.Caption := ..." from code');
  LabelFps.Caption := 'FPS: ' + Container.Fps.ToString;


  ///////////////////////////// the place to change the implementation /////////////////////////////
  LScene := GetSceneUnderMouseWithPixelTest();

  
  if Assigned(LScene) then 
  begin
    if Assigned(FScene) and (FScene <> LScene) then 
      FScene.RenderOptions.WireframeColor := Black.RGB;
    FScene := LScene;
    FScene.RenderOptions.WireframeColor := Red.RGB;
  end
  else
  begin
    if Assigned(FScene) then
      FScene.RenderOptions.WireframeColor := Black.RGB;
    FScene := nil;
  end;
end;

function TViewMain.GetSceneUnderMouse(): TCastleScene;
begin
  Result := Viewport1.TransformUnderMouse as TCastleScene;
end;

function TViewMain.GetSceneUnderMouseWithPixelTest(): TCastleScene;
var
  LRay: TRayCollision;
  LRayNode: TRayCollisionNode;
  LOrigin: TVector3;
  LDirection: TVector3;
  LTexCoords: TVector2;
  LHitLocalCoords: TVector3;
  function IsPixelTransparent(): Boolean;
  var
    LTextureNode: TImageTextureNode;
    LImage: TEncodedImage;
    LPixelPtr: PByte;
    //LPixelColor: TCastleColor;
  begin

    Result := False;    
    if not Assigned(LRayNode.Triangle) then
      Exit;

    // according to comments in the source code, LRayNode.Point is in local coordinates, but the test looks it is in world
    LHitLocalCoords := LRay.Transform.WorldToLocal(LRayNode.Point);
    // according to comments in the source code, ITexCoord2D expects world coordinates, but test shows it takes Local
    LTexCoords := LRayNode.Triangle^.ITexCoord2D(LHitLocalCoords);
    LabelFps.Caption := Format('%s| P %s| TC %s', [LabelFps.Caption, LRayNode.Point.ToString, LTexCoords.ToString]);

    LTextureNode := TImageTextureNode.Create;
    try
      LTextureNode.SetUrl([(LRay.Transform as TCastleScene).Url]);
      LImage := LTextureNode.TextureImage;
      if Assigned(LImage) and Assigned(LImage.RawPixels) and not LImage.IsEmpty then
      begin
        LPixelPtr := PByte(LImage.RawPixels) + Round(LTexCoords.Y * LImage.Height) * LImage.Width * 4 + Round(LTexCoords.X * LImage.Width) * 4;
        //LPixelColor := TCastleColor.Create(LPixelPtr[0] / 255, LPixelPtr[1] / 255, LPixelPtr[2] / 255, LPixelPtr[3] / 255);
        Result := LPixelPtr[3] < 0.5; // Consider pixel transparent if alpha is less than 0.5
      end;
    finally
      FreeAndNil(LTextureNode);
    end;
  end;
begin
  if not Assigned(Viewport1.MouseRayHit) then
    Exit(nil);
  Viewport1.PositionToRay(FMousePos, true, LOrigin, LDirection);
  repeat
    LRay := Viewport1.Items.WorldRay(LOrigin, LDirection);
    if not Assigned(LRay) or not LRay.Info(LRayNode) then
      Exit(nil);
    LabelFps.Caption := Format('%s| MR %s O%s D %s Ray %s RN %s T %s',[LabelFps.Caption, Viewport1.MouseRayHit[0].Point.XY.ToString, 
      '-'{LOrigin.ToString}, LDirection.ToString, Integer(Pointer(LRay)).ToString, Integer(@LRayNode).ToString, Integer(Pointer(LRayNode.Triangle)).ToString]);          
    Result := LRay.Transform as TCastleScene;
    if not IsPixelTransparent() then
      Exit(Result)
    else
    begin
      LOrigin := LRayNode.Point + LDirection.AdjustToLength(SingleEpsilon);
    end; 
  until False;  
end;

// https://forum.castle-engine.io/t/get-all-geometry-under-raycast/957/5
function TViewMain.GetSceneUnderMouseWithPixelTestByBugan(): TCastleScene;
type
  TPairShapeTransform = TPair<TShape, TCastleTransform>;
var
  RayHit: TRayCollision;
  RayOrigin, RayDirection: TVector3;
  Item: TPair<TShape, TCastleTransform>;
  Shape: TShape;
  rayColNode: TRayCollisionNode;
  rayColPos: Integer;
  LHitTransforms: TList<TPairShapeTransform>;
  
  function GetCorrectRayCollisionNodePos(): Integer;
  var
    i: Integer;
  begin
    Result := -1;
    if Assigned(rayHit) then
      for i := rayColPos to rayHit.Count - 1 do
        if Assigned(rayHit.Items[i].Triangle) then
          Exit(i);    
  end;

begin
  if not Assigned(Viewport1.MouseRayHit) then
    Exit(nil);
  Result := nil;
  //depth := Max(depth, 1);
  Viewport1.PositionToRay(Viewport1.MouseRayHit[0].Point.XY, true, RayOrigin, RayDirection);
  RayHit := Viewport1.Items.WorldRay(RayOrigin, RayDirection);
  if RayHit <> nil then
  begin
    //rayColPos := GetCorrectRayCollisionNodePos();
    rayColPos := 0;
    LHitTransforms := TList<TPairShapeTransform>.Create();
    while (RayHit <> nil) {AND (rayColPos > -1)} {AND (depth > 0)} do
    begin
      shape := nil;
      rayColNode := RayHit.Items[rayColPos];
      if (rayColNode.Item is TCastleScene) AND (rayColNode.Triangle <> nil) AND (rayColNode.Triangle.Shape <> nil) then
      begin
        if TCastleScene(rayColNode.Item).VerticesCount <> rayColNode.Triangle.Shape.VerticesCount then
          shape := rayColNode.Triangle.Shape;
      end;

      LHitTransforms.Add(TPair<TShape, TCastleTransform>.Create(shape, rayColNode.Item));
      rayColNode.Item.Pickable := false;
      RayOrigin := rayColNode.Point + RayDirection.AdjustToLength(SingleEpsilon);
      FreeAndNil(RayHit);
      RayHit := Viewport1.Items.WorldRay(RayOrigin, RayDirection);
      Inc(rayColPos);
      //rayColPos := GetCorrectRayCollisionNodePos();
      //Dec(depth);
    end;

    if RayHit <> nil then
      FreeAndNil(RayHit);
    
    for Item in LHitTransforms do
    begin
      Item.Value.Pickable := true;
    end;
    
    if LHitTransforms.Count > 0 then
      Result := LHitTransforms[0].Value as TCastleScene
    else
      Result := nil;
  end;
end;

function TViewMain.Press(const Event: TInputPressRelease): Boolean;
begin
  Result := inherited;
  if Result then Exit; // allow the ancestor to handle keys

  { This virtual method is executed when user presses
    a key, a mouse button, or touches a touch-screen.

    Note that each UI control has also events like OnPress and OnClick.
    These events can be used to handle the "press", if it should do something
    specific when used in that UI control.
    The TViewMain.Press method should be used to handle keys
    not handled in children controls.
  }

  // Use this to handle keys:
  {
  if Event.IsKey(keyXxx) then
  begin
    // DoSomething;
    Exit(true); // key was handled
  end;
  }
end;

end.
